# Initially taken from:
# http://code.activestate.com/recipes/134892/
# Thanks to Danny Yoo
import sys
import tty
import termios


def readchar():
	fd = sys.stdin.fileno()
	old_settings = termios.tcgetattr(fd)
	try:
		tty.setraw(sys.stdin.fileno())
		ch = sys.stdin.read(1)
	finally:
		termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
	return ch
#




















from .Console import Console




class SimpleTableConstants(object):

	HALIGN_LEFT = 0
	HALIGN_CENTER = 1
	HALIGN_RIGHT = 2

	CASE_NORMAL = 0
	CASE_LOWER = 1
	CASE_UPPER = 2

#



class SimpleTableCell(SimpleTableConstants):

	def __init__(self, table):
		self.__table = table
		self.halign = None
		self.value = None
		self.color = None
		self.textTransform = None
	#

	def __str__(self):
		if self.value is None:
			return ""
		else:
			return str(self.value)
	#

	def __len__(self):
		s = self.__str__()
		return len(s)
	#

#



class SimpleTableColumn(SimpleTableConstants):

	def __init__(self, table, columnIndex:int, rows:list):
		self.columnIndex = columnIndex
		self.__table = table
		#self.__columnData = columnData
		self.halign = None
		self.textTransform = None
		self.__rows = rows
		self.color = None
		self.vlineAfterColumn = False
		self.marginLeft = 1
		self.marginRight = 1
	#

	def getMaxWidth(self):
		return max([ (len(row) if row else 0) for row in self.__rows ])
	#

	"""
	@property
	def halign(self) -> int:
		return self.__columnData.get("halign", SimpleTableConstants.HALIGN_DEFAULT)
	#

	@halign.setter
	def halign(self, v:int):
		self.__columnData["halign"] = v
	#
	"""

	def __getitem__(self, index:int):
		assert isinstance(index, int)
		if index >= len(self.__rows):
			return None
		return self.__rows[index]
	#

	def __setitem__(self, index:int, v):
		raise NotImplementedError()
	#

	def __len__(self):
		return len(self.__rows)
	#

	def __enter__(self):
		return self
	#

	def __exit__(self, type, value, traceback):
		pass
	#

#



class SimpleTableRow(SimpleTableConstants):

	def __init__(self, table):
		self.__table = table
		self.__cells = []
		table._getColumnsList(0).append(self)
		self.halign = None
		self.color = None
		self.textTransform = None
		self.hlineAfterRow = False
	#

	def addCell(self) -> SimpleTableCell:
		c = SimpleTableCell(self.__table)
		self.__table._getColumnsList(len(self.__cells)).remove(self)
		self.__cells.append(c)
		self.__table._getColumnsList(len(self.__cells)).append(self)
		return c
	#

	def addCells(self, *args):
		self.__table._getColumnsList(len(self.__cells)).remove(self)
		for a in args:
			c = SimpleTableCell(self.__table)
			self.__cells.append(c)
			c.value = a
		self.__table._getColumnsList(len(self.__cells)).append(self)
	#

	def __getitem__(self, index:int):
		assert isinstance(index, int)
		if index >= len(self.__cells):
			return None
		return self.__cells[index]
	#

	def __setitem__(self, index:int, v):
		raise NotImplementedError()
	#

	def __len__(self):
		return len(self.__cells)
	#

	def __enter__(self):
		return self
	#

	def __exit__(self, type, value, traceback):
		pass
	#

#



class SimpleTable(SimpleTableConstants):

	def __init__(self):
		self.__rows = []
		self.__columns = {}
		self.__nColumns = {
			0: []
		}
	#

	def _getColumnsList(self, n:int):
		cols = self.__nColumns.get(n)
		if cols is None:
			cols = []
			self.__nColumns[n] = cols
		return cols
	#

	def addRow(self, *args) -> SimpleTableRow:
		r = SimpleTableRow(self)
		self.__rows.append(r)
		if args:
			r.addCells(*args)
		return r
	#

	@property
	def numberOfColumns(self) -> int:
		return max(self.__nColumns.keys())
	#

	@property
	def numberOfRows(self) -> int:
		return len(self.__rows)
	#

	def __len__(self):
		return len(self.__rows)
	#

	def __getColumnWidth(self, nColumn:int) -> int:
		cells = self.__getColumnCells(nColumn)
		if cells:
			return max([ (len(c) if c else 0) for c in cells ])
		else:
			return 0
	#

	def __getColumnWidths(self) -> list:
		return [ self.__getColumnWidth(n) for n in range(0, self.numberOfColumns) ]
	#

	def column(self, nColumn:int):
		d = self.__columns.get(nColumn)
		if d is None:
			d = SimpleTableColumn(self, nColumn, self.__rows)
			self.__columns[nColumn] = d
		return d
	#

	def row(self, nRow:int):
		if (nRow >= len(self.__rows)) or (nRow < 0):
			return None
		return self.__rows[nRow]
	#

	def __getColumnCells(self, nColumn:int):
		columnCells = []
		for row in self.__rows:
			columnCells.append(row[nColumn])
		return columnCells
	#

	#
	# Print the table.
	#
	def print(self, prefix:str = "", gapChar = " ", vLineChar = "|", hLineChar = "-", crossChar = "|", printFunction = None, useColors:bool = True):
		if printFunction is None:
			printFunction = print

		outBuffer = []
		self.__printToBuffer(outBuffer, prefix, gapChar, vLineChar, hLineChar, crossChar, useColors)

		for line in outBuffer:
			printFunction(line)
	#

	#
	# Print the table.
	#
	def printToLines(self, prefix:str = "", gapChar = " ", vLineChar = "|", hLineChar = "-", crossChar = "|", useColors:bool = True):
		outBuffer = []
		self.__printToBuffer(outBuffer, prefix, gapChar, vLineChar, hLineChar, crossChar, useColors)
		return outBuffer
	#

	def __printToBuffer(self, outBuffer:list, prefix:str = "", gapChar = " ", vLineChar = "|", hLineChar = "-", crossChar = "|", useColors:bool = True):
		columnWidths = self.__getColumnWidths()

		for row in self.__rows:
			rowCells = [ prefix ]
			data = []
			for nColumn in range(0, len(columnWidths)):
				column = self.column(nColumn)
				halign, color, textTransform, text = self.__getCellData(row, column, row[nColumn])
				if not useColors:
					color = None
				if color:
					rowCells.append(color)
				text = self.__hformatCellText(text, halign, textTransform, columnWidths[nColumn], column.marginLeft, column.marginRight)
				rowCells.append(text)
				if color:
					rowCells.append(Console.RESET)

				if column.vlineAfterColumn:
					rowCells.append(vLineChar)
				else:
					rowCells.append(gapChar)

				data.append((nColumn, column, text))

			outBuffer.append("".join(rowCells))

			if row.hlineAfterRow:
				rowCells.clear()
				rowCells.append(prefix)
				rowGapHLine = hLineChar * len(gapChar)
				for nColumn, column, text in data:
					hline = hLineChar * len(text)
					rowCells.append(hline)

					if column.vlineAfterColumn:
						rowCells.append(crossChar)
					else:
						rowCells.append(rowGapHLine)

				outBuffer.append("".join(rowCells))
	#

	def raw(self):
		ret = []

		for row in self.__rows:
			rowCells = []
			for nColumn in range(0, self.numberOfColumns):
				column = self.column(nColumn)
				halign, color, textTransform, text = self.__getCellData(row, column, row[nColumn])
				rowCells.append(self.__hformatCellText(text, None, textTransform, 0, 0, 0))
			ret.append(rowCells)

		return ret
	#

	def __getCellData(self, row:SimpleTableRow, column:SimpleTableColumn, cell:SimpleTableCell):
		# collect data: cell vs. row vs. column

		if cell:
			s = str(cell)
			halign = cell.halign
			color = cell.color
			textTransform = cell.textTransform
		else:
			s = ""
			halign = None
			color = None
			textTransform = None

		if halign is None:
			halign = row.halign
		if halign is None:
			halign = column.halign

		if color is None:
			color = row.color
		if color is None:
			color = column.color

		if textTransform is None:
			textTransform = row.textTransform
		if textTransform is None:
			textTransform = column.textTransform

		return halign, color, textTransform, s
	#

	def __hformatCellText(self, s:str, halign:int, textTransform:int, width:int, marginLeft:int, marginRight:int):
		if textTransform == SimpleTableConstants.CASE_LOWER:
			s = s.lower()
		elif textTransform == SimpleTableConstants.CASE_UPPER:
			s = s.upper()

		if halign in [ None, SimpleTableConstants.HALIGN_LEFT ]:
			s = s + " " * (width - len(s))
		elif halign == SimpleTableConstants.HALIGN_CENTER:
			spc = " " * ((width - len(s)) // 2 + 1)
			s = spc + s + spc
			s = s[:width]
		elif halign == SimpleTableConstants.HALIGN_RIGHT:
			s = " " * (width - len(s)) + s
		else:
			raise Exception()

		if marginLeft:
			s = " " * marginLeft + s
		if marginRight:
			s += " " * marginRight

		return s
	#

	def addEmptyRow(self):
		if len(self.__rows):
			if (self.__rows[-1]):
				self.addRow()
	#

#











from .Console import Console



class ConsoleGraphics(object):

	@staticmethod
	def hline(x1:int, y:int, x2:int, c:str):
		s = ""
		for i in range(0, x2 - x1 + 1):
			s += c
		Console.printAt(x1, y, s, bFlush=False)
	#

	@staticmethod
	def vline(x:int, y1:int, y2:int, c:str):
		for y in range(y1, y2 + 1):
			Console.printAt(x, y, c, bFlush=False)
	#

	@staticmethod
	def box(centerX:int, centerY:int, text:str, frameChar:str):
		w = len(text)
		x1 = centerX - int(w/2) - 2
		x2 = x1 + 4 + w
		y1 = centerY - 2
		y2 = centerY + 2

		ConsoleGraphics.hline(x1, y1, x2, frameChar)
		ConsoleGraphics.hline(x1, y2, x2, frameChar)
		ConsoleGraphics.vline(x1, y1, y2, frameChar)
		ConsoleGraphics.vline(x2, y1, y2, frameChar)
		Console.printAt(x1 + 2, y1 + 2, text, bFlush=False)
	#

#
















from .Console import *
from .Rect import *






class FramedBoxSettings(object):

	def __init__(self):
		self.bgColor = Console.BackGround.STD_BLACK
		self.bgChar = " "
		self.frameColor = Console.BackGround.STD_BLACK + Console.ForeGround.STD_LIGHTGRAY
		self.frameCharLeft = "|"
		self.frameCharRight = "|"
		self.frameCharTop = "-"
		self.frameCharBottom = "-"
		self.frameCharTopLeft = "/"
		self.frameCharTopRight = "\\"
		self.frameCharBottomLeft = "\\"
		self.frameCharBottomRight = "/"
		self.textColor = Console.BackGround.STD_BLACK + Console.ForeGround.STD_LIGHTGRAY
	#

#




class ConsoleBufferWO(object):

	class _Cell(object):

		def __init__(self, bgColor:str):
			self.col = bgColor
			self.c = " "
			self.bChanged = False
		#

		def clear(self, bgColor:str):
			self.col = bgColor
			self.c = " "
			self.bChanged = True
		#

	#

	def __init__(self, ofsX:int = 0, ofsY:int = 0, width:int = None, height:int = None, bgColor:str = Console.RESET, bAlwaysFullRepaint = False):
		self.__bgColor = bgColor
		self.__width = Console.width() - 1 if width is None else width
		self.__height = Console.height() - 1 if height is None else height
		self.__widthM1 = self.__width - 1
		self.__heightM1 = self.__height - 1
		self.__numberOfCells = self.__width * self.__height
		self.__ofsX = ofsX
		self.__ofsY = ofsY

		self.defaultFramedBoxSettings = FramedBoxSettings()

		self.__lineNumbers = tuple(range(0, self.__height))
		self.__columnNumbers = tuple(range(0, self.__width))

		self.__lastBuffer = [ [ConsoleBufferWO._Cell(bgColor) for x in self.__columnNumbers] for y in self.__lineNumbers]
		self._currentBuffer = [ [ConsoleBufferWO._Cell(bgColor) for x in self.__columnNumbers] for y in self.__lineNumbers]

		assert len(self.__lineNumbers) == self.__height
		assert len(self.__columnNumbers) == self.__width
		assert len(self.__lastBuffer) == self.__height
		assert len(self._currentBuffer) == self.__height
		for i in range(0, self.__height):
			assert len(self.__lastBuffer[i]) == self.__width
			assert len(self._currentBuffer[i]) == self.__width

		self.__bFullPaint = True
		self.__bAlwaysFullRepaint = bAlwaysFullRepaint
	#

	@property
	def bounds(self):
		return Rect(0, 0, self.__width, self.__height)
	#

	@property
	def size(self):
		return self.__width, self.__height
	#

	@property
	def width(self):
		return self.__width
	#

	@property
	def height(self):
		return self.__height
	#

	def clear(self):
		bgColor = self.__bgColor
		for y in self.__lineNumbers:
			for x in self.__columnNumbers:
				self._currentBuffer[y][x].clear(bgColor)
		self.__bFullPaint = True
	#

	def set(self, x:int, y:int, color:str, character:str):
		assert len(character) == 1

		cell = self._currentBuffer[y][x]
		if character != None:
			cell.c = character
			cell.bChanged = True
		if color != None:
			cell.col = color
			cell.bChanged = True
	#

	def hline(self, x1:int, y:int, x2:int, color:str, character:str):
		assert len(character) == 1

		row = self._currentBuffer[y]
		for ix in range(x1, x2 + 1):
			cell = row[ix]
			cell.c = character
			if color != None:
				cell.col = color
			cell.bChanged = True
	#

	def vline(self, x:int, y1:int, y2:int, color:str, character:str):
		assert len(character) == 1

		for iy in range(y1, y2 + 1):
			cell = self._currentBuffer[iy][x]
			cell.c = character
			if color != None:
				cell.col = color
			cell.bChanged = True
	#

	def printBoxedText(self, centerX:int, centerY:int, text:str, settings:FramedBoxSettings = None):
		if settings is None:
			settings = self.defaultFramedBoxSettings

		wtext = len(text)
		x1 = centerX - int(wtext/2) - 2
		x2 = x1 + 4 + wtext
		y1 = centerY - 2
		y2 = centerY + 2
		w = x2 - x1
		h = y2 - y1

		self.hline(x1, y1, x2, settings.frameColor, settings.frameCharTop)
		self.hline(x1, y2, x2, settings.frameColor, settings.frameCharBottom)
		self.vline(x1, y1, y2, settings.frameColor, settings.frameCharLeft)
		self.vline(x2, y1, y2, settings.frameColor, settings.frameCharRight)
		self.set(x1, y1, settings.frameColor, settings.frameCharTopLeft)
		self.set(x2, y1, settings.frameColor, settings.frameCharTopRight)
		self.set(x1, y2, settings.frameColor, settings.frameCharBottomLeft)
		self.set(x2, y2, settings.frameColor, settings.frameCharBottomRight)
		self.fill(x1 + 1, y1 + 1, w - 2, h - 2, settings.bgColor, settings.bgChar)
		self.printText(x1 + 2, y1 + 2, settings.textColor, text)
	#

	def drawFrame(self, rect:Rect = None, x1:int = None, y1:int = None, x2:int = None, y2:int = None, settings:FramedBoxSettings = None):
		if settings is None:
			settings = self.defaultFramedBoxSettings

		if rect:
			x1 = rect.x1
			y1 = rect.y1
			x2 = rect.x2
			y2 = rect.y2

		self.hline(x1, y1, x2, settings.frameColor, settings.frameCharTop)
		self.hline(x1, y2, x2, settings.frameColor, settings.frameCharBottom)
		self.vline(x1, y1, y2, settings.frameColor, settings.frameCharLeft)
		self.vline(x2, y1, y2, settings.frameColor, settings.frameCharRight)
		self.set(x1, y1, settings.frameColor, settings.frameCharTopLeft)
		self.set(x2, y1, settings.frameColor, settings.frameCharTopRight)
		self.set(x1, y2, settings.frameColor, settings.frameCharBottomLeft)
		self.set(x2, y2, settings.frameColor, settings.frameCharBottomRight)
	#

	def fill(self, x:int, y:int, w:int, h:int, color:str, character:str):
		assert len(character) == 1

		for iy in range(y, y + h):
			row = self._currentBuffer[iy]
			for ix in range(x, x + w):
				cell = row[ix]
				cell.c = character
				if color != None:
					cell.col = color
				cell.bChanged = True
	#

	def printText(self, x:int, y:int, color:str, text:str):
		assert len(text) > 0

		i = 0
		for c in text:
			cell = self._currentBuffer[y][x + i]
			cell.c = c
			if color != None:
				cell.col = color
			cell.bChanged = True
			i += 1
	#

	def bufferToConsole(self):
		if self.__bFullPaint or self.__bAlwaysFullRepaint:
			characters = []
			for y in self.__lineNumbers:
				characters.clear()
				lastCol = None

				lastRow = self.__lastBuffer[y]
				currentRow = self._currentBuffer[y]
				for x in self.__columnNumbers:
					lastCell = lastRow[x]
					currentCell = currentRow[x]
					#assert currentCell.col != None
					#assert currentCell.c != None
					lastCell.col = currentCell.col
					lastCell.c = currentCell.c
					if lastCol != currentCell.col:
						lastCol = currentCell.col
						characters.append(lastCol)
					characters.append(currentCell.c)
					currentCell.bChanged = False
				characters.append(Console.RESET)
				Console.printAt(self.__ofsX, self.__ofsY + y, "".join(characters), False)
			self.__bFullPaint = False

		else:
			startX = -1
			characters = []
			bLastWasChanged = False
			lastCol = None

			for y in self.__lineNumbers:
				lastRow = self.__lastBuffer[y]
				currentRow = self._currentBuffer[y]
				for x in self.__columnNumbers:
					lastCell = lastRow[x]
					currentCell = currentRow[x]
					#assert currentCell.col != None
					#assert currentCell.c != None
					if currentCell.bChanged:
						# we might have a change here
						if (currentCell.c != lastCell.c) or (currentCell.col != lastCell.col):
							# yes, we really have a change here
							lastCell.col = currentCell.col
							lastCell.c = currentCell.c
							if bLastWasChanged:
								if lastCol != currentCell.col:
									lastCol = currentCell.col
									characters.append(lastCol)
								characters.append(currentCell.c)
							else:
								if startX >= 0:
									characters.append(Console.RESET)
									Console.printAt(self.__ofsX + startX, self.__ofsY + y, "".join(characters), False)
									characters.clear()
								startX = x
								lastCol = currentCell.col
								characters.append(currentCell.col)
								characters.append(currentCell.c)
								bLastWasChanged = True
						else:
							bLastWasChanged = False
						currentCell.bChanged = False
					else:
						bLastWasChanged = False

				if startX >= 0:
					characters.append(Console.RESET)
					Console.printAt(self.__ofsX + startX, self.__ofsY + y, "".join(characters), False)
					characters.clear()
					startX = -1
					bLastWasChanged = False
					lastCol = None

		print(Console.RESET_TOPLEFT, end="", flush=True)
	#

#









from .Rectangle import Rectangle
from ..Console import Console
from .ViewPortCell import ViewPortCell
from .ViewPortRenderer import ViewPortRenderer
from .ViewPortRGB import ViewPortRGB
from .ViewPortBuffer2 import ViewPortBuffer2









class ConsoleCell(object):

	def __init__(self, viewPortCells:list):
		# contains instances of ViewPortCell
		self.cells = viewPortCells
		# contains the current console string to display
		self.s = None
		# contains the last console string displayed
		self.oldS = None
	#

	def buildConsoleStr(self):
		c = None
		colFG = None
		colFGBrightness = None
		colBG = None
		colBGBrightness = None

		#print(len(self.cells), end="", flush=True)
		for cell in reversed(self.cells):
			if cell.c is not None:
				c = cell.c
				colFG = cell.colFG
				colFGBrightness = cell._brightness
				break

		for cell in reversed(self.cells):
			if cell.colBG is not None:
				colBG = cell.colBG
				colBGBrightness = cell._brightness
				break

		if c is None:
			c = " "
		if colBG is None:
			colBG = 0
			colBGBrightness = 0
		if colFG is None:
			colFG = 0
			colFGBrightness = 0

		self.s = ViewPortRenderer.bgTables[colBGBrightness][colBG] + ViewPortRenderer.fgTables[colFGBrightness][colFG] + c
	#

#




class ViewPort(object):

	__EMPTY_CELL = ViewPortCell()

	def __init__(self):
		ViewPort.__EMPTY_CELL.c = " "
		ViewPort.__EMPTY_CELL.colBG = int(ViewPortRGB.rgb256(0, 0, 0))
		ViewPort.__EMPTY_CELL.colFG = int(ViewPortRGB.rgb256(0, 0, 0))
		ViewPort.__EMPTY_CELL._brightness = ViewPortRenderer.MAX_BRIGHTNESS

		self.__width, self.__height = Console.getSize()
		#self.__height -= 1
		self.__screenMatrix = None
		self.__buffers = []
	#

	def createBuffer(self, *args, bErrorIfOutOfBounds:bool = False):
		if len(args) == 1:
			assert isinstance(args[0], Rectangle)
			ofsX = args[0].x
			ofsY = args[0].y
			width = args[0].width
			height = args[0].height
		elif len(args) == 4:
			assert isinstance(args[0], int)
			assert isinstance(args[1], int)
			assert isinstance(args[2], int)
			assert isinstance(args[3], int)
			ofsX = args[0]
			ofsY = args[1]
			width = args[2]
			height = args[3]
		else:
			raise Exception("args ???")

		assert ofsX >= 0
		assert ofsY >= 0
		assert width >= 0
		assert height >= 0

		buffer = ViewPortBuffer2(self, ofsX, ofsY, width, height)
		buffer.bErrorIfOutOfBounds = bErrorIfOutOfBounds
		self.__buffers.append(buffer)
		self.__screenMatrix = self.__buildScreenMatrix()
		self.__screenStrings = [ None ] * len(self.__screenMatrix)
		self.__oldScreenStrings = [ None ] * len(self.__screenMatrix)
		return buffer
	#

	def __buildScreenMatrix(self):
		rows = []
		for y in range(0, self.__height):
			row = []
			rows.append(row)
			for x in range(0, self.__width):
				row.append(ConsoleCell([ self.__EMPTY_CELL ]))

		for buffer in self.__buffers:
			for y in range(0, buffer.height):
				row = rows[buffer.ofsY + y]
				for x in range(0, buffer.width):
					row[buffer.ofsX + x].cells.append(buffer.rows[y][x])

		return rows
	#

	def render(self, bRenderAll:bool = False):
		bHasData = False
		y = None
		x = None
		output = []

		n = 0
		for nRow, row in enumerate(self.__screenMatrix):
			for nCol, cell in enumerate(row):
				cell.buildConsoleStr()
				if bRenderAll or (cell.s != cell.oldS):
					if not bHasData:
						bHasData = True
						x = nCol
						y = nRow
					output.append(cell.s)
					cell.oldS = cell.s
					n += 1
				else:
					if bHasData:
						Console.printAt(x, y, "".join(output), False)
						bHasData = False
						output.clear()

		if bHasData:
			Console.printAt(x, y, "".join(output), False)
			bHasData = False
		Console.moveCursorTo(0, 0, True)
		#Console.flush()

		return n
	#

#















from typing import Union



class Rectangle(object):

	def __init__(self, *args):
		if len(args) == 0:
			self.__x1 = 0
			self.__y1 = 0
			self.__x2 = 0
			self.__y2 = 0
		elif len(args) == 1:
			arg = args[0]
			if isinstance(arg, (tuple, list)):
				assert len(arg) == 4
				assert isinstance(arg[0], int)
				assert isinstance(arg[1], int)
				assert isinstance(arg[2], int)
				assert isinstance(arg[3], int)
				self.__x1 = arg[0]
				self.__y1 = arg[1]
				self.__x2 = self.__x1 + arg[2]
				self.__y2 = self.__y1 + arg[3]
			elif isinstance(arg, Rectangle):
				self.__x1 = arg.x1
				self.__y1 = arg.y1
				self.__x2 = arg.x2
				self.__y2 = arg.y2
			else:
				raise Exception("arg 0 is " + str(type(arg)) + " ???")
		elif len(args) == 4:
			assert isinstance(args[0], int)
			assert isinstance(args[1], int)
			assert isinstance(args[2], int)
			assert isinstance(args[3], int)
			self.__x1 = args[0]
			self.__y1 = args[1]
			self.__x2 = self.__x1 + args[2]
			self.__y2 = self.__y1 + args[3]
		else:
			raise Exception("args ???")

		self.__w = self.__x2 - self.__x1
		self.__h = self.__y2 - self.__y1
	#

	@property
	def width(self) -> int:
		return self.__w
	#

	@width.setter
	def width(self, value:int):
		assert isinstance(value, int)

		self.__w = value
		self.__x2 = self.__x1 + self.__w
	#

	@property
	def height(self) -> int:
		return self.__h
	#

	@height.setter
	def height(self, value:int):
		assert isinstance(value, int)

		self.__h = value
		self.__y2 = self.__y1 + self.__h
	#

	@property
	def x(self) -> int:
		return self.__x1
	#

	@x.setter
	def x(self, value:int):
		assert isinstance(value, int)

		self.__x1 = value
		self.__w = self.__x2 - self.__x1
	#

	@property
	def y(self) -> int:
		return self.__y1
	#

	@y.setter
	def y(self, value:int):
		assert isinstance(value, int)

		self.__y1 = value
		self.__h = self.__y2 - self.__y1
	#

	@property
	def x1(self) -> int:
		return self.__x1
	#

	@x1.setter
	def x1(self, value:int):
		assert isinstance(value, int)

		self.__x1 = value
		self.__w = self.__x2 - self.__x1
	#

	@property
	def y1(self) -> int:
		return self.__y1
	#

	@y1.setter
	def y1(self, value:int):
		assert isinstance(value, int)

		self.__y1 = value
		self.__h = self.__y2 - self.__y1
	#

	@property
	def x2(self) -> int:
		return self.__x2
	#

	@x2.setter
	def x2(self, value:int):
		assert isinstance(value, int)

		self.__x2 = value
		self.__w = self.__x2 - self.__x1
	#

	@property
	def y2(self) -> int:
		return self.__y2
	#

	@y2.setter
	def y2(self, value:int):
		assert isinstance(value, int)

		self.__y2 = value
		self.__h = self.__y2 - self.__y1
	#

	@property
	def topLeft(self) -> tuple:
		return (self.__x1, self.__y1)
	#

	@topLeft.setter
	def topLeft(self, value:Union[tuple, list]):
		assert isinstance(value, (tuple, list))
		assert len(value) == 2

		self.__x1 = value[0]
		self.__y1 = value[1]
		self.__w = self.__x2 - self.__x1
		self.__h = self.__y2 - self.__y1
	#

	@property
	def topRight(self) -> tuple:
		return (self.__x2 - 1, self.__y1)
	#

	@topRight.setter
	def topRight(self, value:Union[tuple, list]):
		assert isinstance(value, (tuple, list))
		assert len(value) == 2

		self.__x2 = value[0] + 1
		self.__y1 = value[1]
		self.__w = self.__x2 - self.__x1
		self.__h = self.__y2 - self.__y1
	#

	@property
	def bottomRight(self) -> tuple:
		return (self.__x2 - 1, self.__y2 - 1)
	#

	@bottomRight.setter
	def bottomRight(self, value:Union[tuple, list]):
		assert isinstance(value, (tuple, list))
		assert len(value) == 2

		self.__x2 = value[0] + 1
		self.__y2 = value[1] + 1
		self.__w = self.__x2 - self.__x1
		self.__h = self.__y2 - self.__y1
	#

	@property
	def bottomLeft(self) -> tuple:
		return (self.__x1, self.__y2 - 1)
	#

	@bottomLeft.setter
	def bottomLeft(self, value:Union[tuple, list]):
		assert isinstance(value, (tuple, list))
		assert len(value) == 2

		self.__x1 = value[0]
		self.__y2 = value[1] + 1
		self.__w = self.__x2 - self.__x1
		self.__h = self.__y2 - self.__y1
	#

	def isValid(self) -> bool:
		return (self.__w > 0) and (self.__h > 0)
	#

	def area(self) -> int:
		return self.__w * self.__h
	#

	#def clone(self) -> Rectangle:
	def clone(self):
		return Rectangle(self)
	#

	def enlarge(self, *args):
		if len(args) == 1:
			v = args[0]
			if isinstance(v, Rectangle):
				self.__x1 -= v.x1
				self.__y1 -= v.y1
				self.__x2 += v.x2
				self.__y2 += v.y2
			else:
				self.__x1 -= v
				self.__y1 -= v
				self.__x2 += v
				self.__y2 += v
		elif len(args) == 2:
			vh = args[0]
			vv = args[1]
			self.__x1 -= vh
			self.__y1 -= vv
			self.__x2 += vh
			self.__y2 += vv
		elif len(args) == 4:
			self.__x1 -= args[0]
			self.__y1 -= args[1]
			self.__x2 += args[2]
			self.__y2 += args[3]
		else:
			raise Exception("args ???")

		self.__w = self.__x2 - self.__x1
		self.__h = self.__y2 - self.__y1

		return self
	#

	def shrink(self, *args):
		if len(args) == 1:
			v = args[0]
			if isinstance(v, Rectangle):
				self.__x1 += v.x1
				self.__y1 += v.y1
				self.__x2 -= v.x2
				self.__y2 -= v.y2
			else:
				self.__x1 += v
				self.__y1 += v
				self.__x2 -= v
				self.__y2 -= v
		elif len(args) == 2:
			vh = args[0]
			vv = args[1]
			self.__x1 += vh
			self.__y1 += vv
			self.__x2 -= vh
			self.__y2 -= vv
		elif len(args) == 4:
			self.__x1 += args[0]
			self.__y1 += args[1]
			self.__x2 -= args[2]
			self.__y2 -= args[3]
		else:
			raise Exception("args ???")

		self.__w = self.__x2 - self.__x1
		self.__h = self.__y2 - self.__y1

		return self
	#

	#def intersect(self, other:Rectangle) -> Rectangle:
	def intersect(self, other):
		assert isinstance(other, Rectangle)

		if (other.__x1 > self.__x2) \
			or (other.__y1 > self.__y2) \
			or (other.__x2 < self.__x1) \
			or (other.__y2 < self.__y1):
			# no intersection
			return None

		x1 = max(self.__x1, other.__x1)
		y1 = max(self.__y1, other.__y1)
		x2 = min(self.__x2, other.__x2)
		y2 = min(self.__y2, other.__y2)

		return Rectangle(x1, y2, x2 - x1, y2 - y1)
	#

	#def unite(self, other:Rectangle) -> Rectangle:
	def unite(self, other):
		assert isinstance(other, Rectangle)

		x1 = min(self.__x1, other.__x1)
		y1 = min(self.__y1, other.__y1)
		x2 = max(self.__x2, other.__x2)
		y2 = max(self.__y2, other.__y2)

		return Rectangle(x1, y1, x2 - x1, y2 - y1)
	#

	@staticmethod
	#def intersectMany(other) -> Rectangle:
	def intersectMany(other):
		assert isinstance(other, (tuple, list))

		if len(other) == 0:
			raise Exception("args ???")

		if len(other) == 1:
			assert isinstance(other, Rectangle)
			return other

		rect = other[0]
		assert isinstance(rect, Rectangle)
		x1 = rect.__x1
		y1 = rect.__y1
		x2 = rect.__x2
		y2 = rect.__y2

		for r in other[1:]:
			assert isinstance(r, Rectangle)

			if (r.__x1 > x2) \
				or (r.__y1 > y2) \
				or (r.__x2 < x1) \
				or (r.__y2 < y1):
				# no intersection
				return None

			x1 = max(x1, r.__x1)
			y1 = max(y1, r.__y1)
			x2 = min(x2, r.__x2)
			y2 = min(y2, r.__y2)

		return Rectangle(x1, y1, x2 - x1, y2 - y1)
	#

	@staticmethod
	#def uniteMany(other) -> Rectangle:
	def uniteMany(other):
		assert isinstance(other, (tuple, list))

		x1 = min([ r.__x1 for r in other ])
		y1 = min([ r.__y1 for r in other ])
		x2 = max([ r.__x2 for r in other ])
		y2 = max([ r.__y2 for r in other ])

		return Rectangle(x1, y2, x2 - x2, y2 - y1)
	#

	def shift(self, x:int, y:int):
		assert isinstance(x, int)
		assert isinstance(y, int)

		self.__x1 += x
		self.__x2 += x
		self.__y1 += y
		self.__y2 += y
	#

#












from .Rect import Rect
from .Console import Console
from .ConsoleGraphics import ConsoleGraphics	# console buffer: cells with color ASCII string, text character and modification-flag; the buffer itself is without offset;
from .ConsoleBuffer import ConsoleBuffer		# console buffer: cells with color ASCII string, text character and modification-flag; the buffer itself has an offset value stored internall;
from .ConsoleBufferWO import ConsoleBufferWO	# simple version of ConsoleBuffer without modification flag
from .CharacterBuffer import CharacterBuffer	# int-based RGB values
from .IntRGB import IntRGB
from .ConsoleBufferRGB import ConsoleBufferRGB
from .SimpleTable import SimpleTable, SimpleTableCell, SimpleTableColumn, SimpleTableRow, SimpleTableConstants
from .EventCollector import EventCollector








__version__ = "0.2019.10.7.1"







class Rect(object):

	def __init__(self, x:int, y:int, w:int, h:int):
		assert w >= 0
		assert h >= 0
		self._x1 = x
		self._y1 = y
		self._w = w
		self._h = h
		self._x2 = x + w
		self._y2 = y + h
	#

	@property
	def x(self):
		return self._x1
	#

	@property
	def y(self):
		return self._y1
	#

	@property
	def w(self):
		return self._w
	#

	@property
	def width(self):
		return self._w
	#

	@property
	def h(self):
		return self._h
	#

	@property
	def height(self):
		return self._h
	#

	@property
	def x1(self):
		return self._x1
	#

	@property
	def y1(self):
		return self._y1
	#

	@property
	def x2(self):
		return self._x2
	#

	@property
	def y2(self):
		return self._y2
	#

	def __str__(self):
		return "Rect(" + str(self._x1) + ", " + str(self._y1) + ", " + str(self._x2) + ", " + str(self._y2) + ")"

	def move(self, x:int, y:int):
		return Rect(self._x1 + x, self._y1 + x, self._w, self._h)
	#

	def setPosition(self, x:int, y:int):
		return Rect(x, x, self._w, self._h)
	#

	def setSize(self, w:int, h:int):
		return Rect(self._x1, self._y1, w, h)
	#

	def isPointInRect(self, x:int, y:int):
		return (x >= self._x1) and (y >= self._y1) and (x <= self._x2) and (y <= self._y2)
	#

	def isOtherRectInRect(self, other):
		assert isinstance(other, Rect)
		return (other._x1 >= self._x1) and (other._y1 >= self._y1) and (other._x2 <= self._x2) and (other._y2 <= self._y2)
	#

	def shrink(self, left:int = 0, bottom:int = 0, right:int = 0, top:int = 0, all:int = 0):
		if all != 0:
			x1 = self._x1 + all
			x2 = self._x2 - all
			y1 = self._y1 + all
			y2 = self._y2 - all
		else:
			x1 = self._x1 + left
			x2 = self._x2 - right
			y1 = self._y1 + top
			y2 = self._y2 - bottom
		return Rect(x1, y1, x2 - x1, y2 - y1)
	#

	def grow(self, left:int = 0, bottom:int = 0, right:int = 0, top:int = 0, all:int = 0):
		if all != 0:
			x1 = self._x1 - all
			x2 = self._x2 + all
			y1 = self._y1 - all
			y2 = self._y2 + all
		else:
			x1 = self._x1 - left
			x2 = self._x2 + right
			y1 = self._y1 - top
			y2 = self._y2 + bottom
		return Rect(x1, y1, x2 - x1, y2 - y1)
	#

	def union(self, other):
		assert isinstance(other, Rect)
		x1 = min(self._x1, other._x1, self._x2, other._x2)
		y1 = min(self._y1, other._y1, self._y2, other._y2)
		x2 = max(self._x1, other._x1, self._x2, other._x2)
		y2 = max(self._y1, other._y1, self._y2, other._y2)
		w = x2 - x1
		h = y2 - y1
		if (w < 0) or (h < 0):
			return None
		else:
			return Rect(x1, y1, w, h)
	#

	def intersect(self, other):
		assert isinstance(other, Rect)
		if (other._x2 < self._x1) or (other._y2 < self._y1) \
			or (self._x2 < other._x1) or (self._y2 < other._y1):
			return None
		x1 = other._x1 if self._x1 < other._x1 else self._x1
		y1 = other._y1 if self._y1 < other._y1 else self._y1
		x2 = other._x2 if self._x2 > other._x2 else self._x2
		y2 = other._y2 if self._y2 > other._y2 else self._y2
		w = x2 - x1
		h = y2 - y1
		return Rect(x1, y1, w, h)
	#

#














import sys
import threading

from .Console import Console





#bPynputLoaded = False
#try:
#	import pynput
#	from pynput.mouse import Listener as MouseListener
#	bPynputLoaded = True
#except:
#	pass





#
# This class implements an active component that actively retrieves keyboard or mouse events. Whenever an event is fired an event handler
# method will be invoked (if such a method was specified).
#
class EventCollector(object):

	def __getCallable(self, attrName:str, obj, function):
		if obj:
			try:
				f = getattr(obj, attrName)
			except:
				f = None
			if f:
				assert callable(f)
		else:
			if function:
				assert callable(function)
			f = function
		return f
	#

	#def __xx_on_mouse_move(self, x, y):
	#	self.__on_mouse_move(self, x, y)
	##

	def __init__(self, listener = None, on_key_pressed = None, on_ctrl_c = None, on_screen_resized = None,
		on_mouse_move = None, on_mouse_wheel = None, on_mouse_button = None, on_mouse_drag = None):

		self.__on_screen_resized = self.__getCallable("on_screen_resized", listener, on_screen_resized)
		self.__on_key_pressed = self.__getCallable("on_key_pressed", listener, on_key_pressed)
		self.__on_ctrl_c = self.__getCallable("on_ctrl_c", listener, on_ctrl_c)
		self.__on_mouse_move = self.__getCallable("on_mouse_move", listener, on_mouse_move)
		self.__on_mouse_wheel = self.__getCallable("on_mouse_wheel", listener, on_mouse_wheel)
		self.__on_mouse_button = self.__getCallable("on_mouse_button", listener, on_mouse_button)
		self.__on_mouse_drag = self.__getCallable("on_mouse_drag", listener, on_mouse_drag)

		#if bPynputLoaded:
		#	self.__mouseListener = MouseListener(
		#		on_move = self.__xx_on_mouse_move if self.__on_mouse_move else None,
		#		)
		#	self.__mouseListener.start()
		#else:
		#	self.__mouseListener = None

		self.__bRun = True
	#

	def terminate(self):
		self.__bRun = False
		#if self.__mouseListener:
		#	self.__mouseListener.stop()
	#

	def run(self):
		print(Console.MOUSE_ENABLE)
		mouseButtonStates = [ False ] * 3
		try:

			lastSize = Console.getSize()
			while self.__bRun:
				key = Console.Input.readKeyWithTimeout(1)

				currentSize = Console.getSize()
				if lastSize != currentSize:
					if self.__on_screen_resized:
						self.__on_screen_resized(self, lastSize[0], lastSize[1], currentSize[0], currentSize[1])
					lastSize = currentSize

				if key is None:
					pass
				elif key.startswith(Console.Input.MOUSE_EVENT):
					bMouseDown = key[-1] == "M"
					nMouseButton, mouseX, mouseY = [ int(x) for x in key[3:-1].split(";") ]
					if nMouseButton == 64:
						if self.__on_mouse_wheel:
							self.__on_mouse_wheel(self, False)
					elif nMouseButton == 65:
						if self.__on_mouse_wheel:
							self.__on_mouse_wheel(self, True)
					else:
						if 0 <= nMouseButton <= 2:
							bDragging = bMouseDown and (mouseButtonStates[nMouseButton] == bMouseDown)
							mouseButtonStates[nMouseButton] = bMouseDown
							if bDragging:
								if self.__on_mouse_drag:
									self.__on_mouse_drag(self, nMouseButton, mouseX, mouseY, tuple(mouseButtonStates))
							else:
								if self.__on_mouse_button:
									self.__on_mouse_button(self, nMouseButton, bMouseDown, mouseX, mouseY, tuple(mouseButtonStates))
					# print(bMouseDown, nMouseButton, mouseX, mouseY)
				elif key is Console.Input.KEY_CTRL_C:
					if self.__on_ctrl_c:
						self.__on_ctrl_c(self, key)
					else:
						self.__bRun = False
				else:
					if self.__on_key_pressed:
						self.__on_key_pressed(self, key)

		finally:
			print(Console.MOUSE_DISABLE)
	#

#





















from .Console import *
from .Rect import *






class FramedBoxSettings(object):

	def __init__(self):
		self.bgColor = Console.BackGround.STD_BLACK
		self.bgChar = " "
		self.frameColor = Console.BackGround.STD_BLACK + Console.ForeGround.STD_LIGHTGRAY
		self.frameCharLeft = "|"
		self.frameCharRight = "|"
		self.frameCharTop = "-"
		self.frameCharBottom = "-"
		self.frameCharTopLeft = "/"
		self.frameCharTopRight = "\\"
		self.frameCharBottomLeft = "\\"
		self.frameCharBottomRight = "/"
		self.textColor = Console.BackGround.STD_BLACK + Console.ForeGround.STD_LIGHTGRAY
	#

#




class ConsoleBuffer_Mixins:

	def _Mixins_init(self):
		self.defaultFramedBoxSettings = FramedBoxSettings()
	#

	#
	# Fill the entire buffer with the default background color.
	#
	def clear(self):
		bgColor = self._bgColor
		for y in self._lineNumbers:
			for x in self._columnNumbers:
				self._data[y][x].clear(bgColor)
		self._bFullPaint = True
	#

	#
	# Fill the entire buffer with the specified color and character.
	#
	def fill(self, color:str, character:str):
		if character != None:
			assert len(character) == 1

		for y in self._lineNumbers:
			for x in self._columnNumbers:
				cell = self._data[y][x]
				if character != None:
					cell.c = character
				if color != None:
					cell.col = color
		self._bFullPaint = True
	#

	#
	# Fill a single cell with color and character.
	#
	def set(self, x:int, y:int, color:str, character:str):
		cell = self._data[y][x]
		if character != None:
			assert len(character) == 1
			cell.c = character
			cell.bChanged = True
		if color != None:
			cell.col = color
			cell.bChanged = True
	#

	#
	# Fill a single cell with color and character.
	#
	def setSafe(self, x:int, y:int, color:str, character:str):
		if (x < 0) or (y < 0) or (x >= self._width) or (y >= self._height):
			return
		cell = self._data[y][x]
		if character != None:
			assert len(character) == 1
			cell.c = character
			cell.bChanged = True
		if color != None:
			cell.col = color
			cell.bChanged = True
	#

	def printText(self, x:int, y:int, color:str, text:str):
		assert len(text) > 0

		i = 0
		row = self._data[y]
		for c in text:
			if (x + i) >= len(row):
				break
			cell = row[x + i]
			cell.c = c
			if color != None:
				cell.col = color
			cell.bChanged = True
			i += 1
	#

	def hline(self, x1:int, y:int, x2:int, color:str, character:str):
		assert len(character) == 1

		row = self._data[y]
		for ix in range(x1, x2 + 1):
			cell = row[ix]
			cell.c = character
			if color != None:
				cell.col = color
			cell.bChanged = True
	#

	def vline(self, x:int, y1:int, y2:int, color:str, character:str):
		assert len(character) == 1

		for iy in range(y1, y2 + 1):
			cell = self._data[iy][x]
			cell.c = character
			if color != None:
				cell.col = color
			cell.bChanged = True
	#

	def printBoxedText(self, centerX:int, centerY:int, text:str, settings:FramedBoxSettings = None):
		if settings is None:
			settings = self.defaultFramedBoxSettings

		wtext = len(text)
		x1 = centerX - int(wtext/2) - 2
		x2 = x1 + 4 + wtext
		y1 = centerY - 2
		y2 = centerY + 2
		w = x2 - x1
		h = y2 - y1

		self.hline(x1, y1, x2, settings.frameColor, settings.frameCharTop)
		self.hline(x1, y2, x2, settings.frameColor, settings.frameCharBottom)
		self.vline(x1, y1, y2, settings.frameColor, settings.frameCharLeft)
		self.vline(x2, y1, y2, settings.frameColor, settings.frameCharRight)
		self.set(x1, y1, settings.frameColor, settings.frameCharTopLeft)
		self.set(x2, y1, settings.frameColor, settings.frameCharTopRight)
		self.set(x1, y2, settings.frameColor, settings.frameCharBottomLeft)
		self.set(x2, y2, settings.frameColor, settings.frameCharBottomRight)
		self.fill(x1 + 1, y1 + 1, w - 2, h - 2, settings.bgColor, settings.bgChar)
		self.printText(x1 + 2, y1 + 2, settings.textColor, text)
	#

	def drawFrame(self, rect:Rect = None, x1:int = None, y1:int = None, x2:int = None, y2:int = None, settings:FramedBoxSettings = None):
		if settings is None:
			settings = self.defaultFramedBoxSettings

		if rect:
			x1 = rect.x1
			y1 = rect.y1
			x2 = rect.x2
			y2 = rect.y2

		self.hline(x1, y1, x2, settings.frameColor, settings.frameCharTop)
		self.hline(x1, y2, x2, settings.frameColor, settings.frameCharBottom)
		self.vline(x1, y1, y2, settings.frameColor, settings.frameCharLeft)
		self.vline(x2, y1, y2, settings.frameColor, settings.frameCharRight)
		self.set(x1, y1, settings.frameColor, settings.frameCharTopLeft)
		self.set(x2, y1, settings.frameColor, settings.frameCharTopRight)
		self.set(x1, y2, settings.frameColor, settings.frameCharBottomLeft)
		self.set(x2, y2, settings.frameColor, settings.frameCharBottomRight)
	#

	def fillRectangle(self, x:int, y:int, w:int, h:int, color:str, character:str):
		assert len(character) == 1

		for iy in range(y, y + h):
			row = self._data[iy]
			for ix in range(x, x + w):
				cell = row[ix]
				cell.c = character
				if color != None:
					cell.col = color
				cell.bChanged = True
	#

#









from .ViewPortRGB import ViewPortRGB
from .ViewPortRenderer import ViewPortRenderer





class ViewPortCell(object):

	def __init__(self, c:str = None, colBG:ViewPortRGB = None, colFG:ViewPortRGB = None):
		self.c = c
		self.colBG = None if colBG is None else int(colBG)
		self.colFG = None if colFG is None else int(colFG)
		self._brightness = ViewPortRenderer.MAX_BRIGHTNESS
	#

	def set(self, c:str, colBG:ViewPortRGB, colFG:ViewPortRGB):
		if c is not None:
			self.c = c
		if colBG is not None:
			self.colBG = int(colBG)
		if colFG is not None:
			self.colFG = int(colFG)
	#

#








#
# This example uses a console buffer implementation to perform fast rendering of text.
# Performance measurements are displayed.
# Performance might vary greatly from system to system.
#



import sys
import time
import datetime

from jk_console import *





class Effect1(object):

	def __init__(self, width:int = None, height:int = None):
		self.chars = ".,;/;,. "
		self.colors = [ Console.ForeGround.STD_LIGHTGRAY, Console.ForeGround.STD_BLUE, Console.ForeGround.STD_LIGHTBLUE, Console.ForeGround.STD_LIGHTCYAN,
			Console.ForeGround.STD_LIGHTBLUE, Console.ForeGround.STD_BLUE, Console.ForeGround.STD_LIGHTGRAY, Console.RESET ]
		self.mods = [ 0, -1, -1, -2, -2, -1, -1, 0 ]

		self.expectedWidth = width
		self.expectedHeight = height
		self.w = None
		self.h = None
		self.cb = None
		self.i = None
		self.lastConsoleWidth = -1
		self.lastConsoleHeight = -1

		self.init()
	#

	def init(self):
		self.lastConsoleWidth, self.lastConsoleHeight = Console.getSize()

		self.w = Console.width() - 1 if self.expectedWidth is None else min(self.expectedWidth, Console.width() - 1)
		self.h = Console.height() - 1 if self.expectedHeight is None else min(self.expectedHeight, Console.height() - 1)
		self.cb = ConsoleBuffer(self.w, self.h)

		rect = Rect(0, 0, self.w, self.h)
		rect = rect.shrink(right = 1, bottom = 1)
		self.cb.drawFrame(rect = rect)

		self.i = 1
	#

	def drawPattern(self, ofs:int):
		for iy in range(1, self.h - 2):
			for ix in range(2, self.w - 2):
				j = (iy + ofs) % 8
				i = (ix + iy + ofs + self.mods[j]) % 8
				self.cb.set(ix, iy, self.colors[i], self.chars[i])
	#

	def runOneLoop(self):
		t0 = datetime.datetime.now()
		self.drawPattern(self.i)
		t1 = datetime.datetime.now()
		self.cb.bufferToConsole(bForceFullRepaint=True)
		t2 = datetime.datetime.now()

		#time.sleep(0.001)

		td1 = t1 - t0
		td2 = t2 - t1
		tdS = t2 - t0
		pdata = [
			"~~ screen-size: ", str(self.w), "x", str(self.h), " pixels"
			" ~~ rendering: %d.%03ds" % (td1.seconds, (td1.microseconds + 499) // 1000),
			" ~~ buffer-to-screen: %d.%03ds" % (td2.seconds, (td2.microseconds + 499) // 1000),
			" ~~ total: %d.%03ds (fps %d)" % (tdS.seconds, (tdS.microseconds + 499) // 1000, int(1.0 / (t2 - t0).total_seconds())),
		]
		pdata.append(" ~~")
		self.cb.printText(2, self.h - 2, Console.ForeGround.STD_WHITE, "".join(pdata))
		self.i += 1

		ww, hh = Console.getSize()
		if (ww != self.lastConsoleWidth) or (hh != self.lastConsoleHeight):
			self.init()
	#

	def runForever(self):
		while True:
			self.runOneLoop()
	#

#


